from hashlib import md5
import os
from pathlib import Path
from typing import Mapping, Optional
from kgdata.wikidata.models import WDClass, WDProperty


ROOT_DIR = Path(__file__).parent.parent.absolute()
DB_DIR = (
    Path(__file__).parent.parent.parent / "data/home/databases"
    if os.environ.get("DB_DIR") is None
    else Path(os.environ["DB_DIR"])
)
assert DB_DIR.exists()


def get_class_readable_label(
    eid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
):
    if wdclasses is None:
        return None

    if wdredirections is not None and eid in wdredirections:
        eid = wdredirections[eid]

    return f"{wdclasses[eid].label} ({eid})"


def get_prop_readable_label(
    pid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    if wdprops is None:
        return None

    if wdredirections is not None and pid in wdredirections:
        pid = wdredirections[pid]

    return f"{wdprops[pid].label} ({pid})"



"""NOTE: this script is copied from a local folder and hasn't been re-tested. For now it serves as
a reference how semtab2020 dataset is converted
"""

import csv
from collections import defaultdict
from io import StringIO
from pathlib import Path
from typing import List, Mapping, Optional, Tuple, Union, cast
from zipfile import ZipFile

import orjson
import serde.csv
import serde.textline
from kgdata.wikidata.db import get_class_db, get_entity_redirection_db, get_prop_db
from kgdata.wikidata.models import WDClass, WDProperty
from loguru import logger
from rdflib import RDFS
from sm.dataset import FullTable
from sm.inputs.link import WIKIDATA, EntityId
from sm.misc.matrix import Matrix
from sm.namespaces.wikidata import WikidataNamespace
from sm.prelude import I, M, O
from tqdm import tqdm

from grams.inputs.linked_table import Context, Link, LinkedTable
from scripts.config import DATA_DIR, DATASET_DIR

PathOrStr = Union[str, Path]


def normalize_semtab2022(
    indir: PathOrStr,
    outdir: PathOrStr,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    """Normalize semtab2022 dataset. This function tested on HardTablesR1.

    The original semtab2022 dataset should have the following format:

    The output will have the following format:
    descriptions
    ├── part-<num>.zip
    │   ├── <table_file_name>.json
    tables
    ├── part-<num>.zip
    │   ├── <table_file_name>.json

    Args:
        indir: the original Round_<num> directory
        outdir: the output directory
    """
    indir = Path(indir) / "Test"
    outdir = Path(outdir)

    # round number
    cea_gt = indir / f"gt/cea_gt.csv"
    cpa_gt = indir / f"gt/cpa_gt.csv"
    cta_gt = indir / f"gt/cta_gt.csv"

    cea = serde.csv.deser(cea_gt)
    cpa = serde.csv.deser(cpa_gt)
    cta = serde.csv.deser(cta_gt)

    # gather links and prop/types
    table2links = defaultdict(list)
    table2sem = defaultdict(lambda: ([], []))

    for row in cea:
        table2links[row[0]].append(row)
    for row in cpa:
        table2sem[row[0]][0].append(row)
    for row in cta:
        table2sem[row[0]][1].append(row)

    wdns = WikidataNamespace.create()
    outputs: List[Tuple[O.SemanticModel, FullTable]] = []

    cea_tables = {row[0] for row in cea}
    cta_tables = {row[0] for row in cta}

    # verifying that the target and ground truth are the same
    for i, row in enumerate(serde.csv.deser(indir / f"target/cea_target.csv")):
        if i >= len(cea) or row != cea[i][:3]:
            assert (
                row[0] not in cea_tables
            ), f"when a target entity is not in ground-truth. the whole table {row[0]} must not be in the ground-truth"
    for i, row in enumerate(serde.csv.deser(indir / f"target/cpa_target.csv")):
        assert row == cpa[i][:3]
    for i, row in enumerate(serde.csv.deser(indir / f"target/cta_target.csv")):
        if i >= len(cta) or row != cta[i][:2]:
            assert (
                row[0] not in cta_tables
            ), f"when a target column is not in ground-truth. the whole table {row[0]} must not be in the ground-truth"

    print("Number of tables:")
    print("- CEA ground-truth:", len(cea_tables))
    print("- CTA ground-truth:", len(set(row[0] for row in cta)))
    print("- CPA ground-truth:", len(cta_tables))

    # extract tables
    print("- # CSV File:", sum(1 for _ in (indir / "tables").glob("*.csv")))
    ignore_logs = []

    for file in tqdm(sorted((indir / "tables").glob("*.csv"))):
        table_id = file.stem
        with open(file, mode="r") as f:
            reader = csv.reader(f, delimiter=",")
            lst = [row for row in reader]
            header, rows = lst[0], lst[1:]
            columns = []
            for ci in range(len(header)):
                columns.append(
                    I.Column(
                        index=ci, name=header[ci], values=[row[ci] for row in rows]
                    )
                )

            table = I.ColumnBasedTable(table_id=table_id, columns=columns)
            shp = table.shape()
            links = [[[] for ci in range(shp[1])] for ri in range(shp[0])]
            for row in table2links[table_id]:
                ri, ci = int(row[1]) - 1, int(row[2])
                ent_uri = row[-1]
                links[ri][ci].append(
                    Link(
                        start=0,
                        end=len(table[ri, ci]),
                        url=None,
                        entities=[EntityId(wdns.get_entity_id(ent_uri), WIKIDATA)],
                    )
                )

            sm = O.SemanticModel()
            col2node = {}
            for col in table.columns:
                col2node[col.index] = sm.add_node(
                    O.DataNode(col.index, cast(str, col.name))
                )

            rels, types = table2sem[table_id]
            ignore_table = False

            for row in types:
                cid = int(row[1])
                ent_id = wdns.get_entity_id(row[-1])

                try:
                    label = get_class_readable_label(ent_id, wdredirections, wdclasses)
                except KeyError:
                    logger.error(
                        "Table {} contains unknown class {} in column {}",
                        table_id,
                        ent_id,
                        cid,
                    )
                    label = None
                    if wdclasses is not None and wdredirections is not None:
                        ignore_table = True
                        ignore_logs.append(
                            f"Ignore table {table_id} as it contains unknown class {ent_id} in column {cid}"
                        )
                        break

                class_id = sm.add_node(
                    O.ClassNode(
                        abs_uri=wdns.get_entity_abs_uri(ent_id),
                        rel_uri=wdns.get_entity_rel_uri(ent_id),
                        approximation=False,
                        readable_label=label,
                    )
                )
                sm.add_edge(
                    O.Edge(
                        source=class_id,
                        target=col2node[cid],
                        abs_uri=str(RDFS.label),
                        rel_uri=wdns.get_rel_uri(RDFS.label),
                        approximation=False,
                        readable_label=wdns.get_rel_uri(RDFS.label),
                    )
                )
                col2node[cid] = class_id

            if ignore_table:
                continue

            for row in rels:
                pid = wdns.get_prop_id(row[-1])
                c1, c2 = int(row[1]), int(row[2])
                try:
                    label = get_prop_readable_label(pid, wdredirections, wdprops)
                except KeyError:
                    logger.error(
                        "Table {} contains unknown property {} from column {} -> {}",
                        table_id,
                        pid,
                        c1,
                        c2,
                    )
                    label = None
                    if wdprops is not None and wdredirections is not None:
                        ignore_table = True
                        ignore_logs.append(
                            f"Ignore table {table_id} as it contains unknown property {pid} from column {c1} -> {c2}"
                        )
                        break

                source_node = sm.get_node(col2node[c1])
                if not (
                    isinstance(source_node, O.ClassNode)
                    or (
                        isinstance(source_node, O.LiteralNode)
                        and source_node.datatype == O.LiteralNodeDataType.Entity
                    )
                ):
                    ignore_table = True
                    ignore_logs.append(
                        f"Ignore table {table_id} as it contains edge from column {c1} -> {c2} but {c1} is not a class or entity node"
                    )
                    break
                sm.add_edge(
                    O.Edge(
                        source=col2node[c1],
                        target=col2node[c2],
                        abs_uri=wdns.get_prop_abs_uri(pid),
                        rel_uri=wdns.get_prop_rel_uri(pid),
                        approximation=False,
                        readable_label=label,
                    )
                )

            if ignore_table:
                continue

            outputs.append(
                (sm, FullTable(table=table, context=Context(), links=Matrix(links)))
            )

    serde.textline.ser(ignore_logs, outdir / "ignore_tables.log")

    outputs = sorted(outputs, key=lambda x: x[1].table.table_id)
    (outdir / "descriptions").mkdir(exist_ok=True, parents=True)
    (outdir / "tables").mkdir(exist_ok=True, parents=True)

    batch_size = 500

    for name in ["CEA", "CPA", "CTA"]:
        # compress the content to save space, use zless, zcat to browse it
        content = (indir / f"target/{name.lower()}_target.csv").read_bytes()
        (outdir / f"{name}_targets.csv").write_bytes(content)

    counter = 0
    for i in range(0, len(outputs), batch_size):
        batch = outputs[i : i + batch_size]
        filename = f"part-{counter:04d}.zip"
        with ZipFile(outdir / "descriptions" / filename, "w") as zf:
            for sm, table in batch:
                zf.writestr(
                    table.table.table_id + ".json", data=orjson.dumps([sm.to_dict()])
                )
        with ZipFile(outdir / "tables" / filename, "w") as zf:
            for sm, table in batch:
                zf.writestr(
                    table.table.table_id + ".json", data=orjson.dumps(table.to_dict())
                )
        counter += 1


def get_prop_readable_label(
    pid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    if wdprops is None:
        return None

    if wdredirections is not None and pid in wdredirections:
        pid = wdredirections[pid]

    return f"{wdprops[pid].label} ({pid})"


def get_class_readable_label(
    eid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
):
    if wdclasses is None:
        return None

    if wdredirections is not None and eid in wdredirections:
        eid = wdredirections[eid]

    return f"{wdclasses[eid].label} ({eid})"


if __name__ == "__main__":
    wdredirections = get_entity_redirection_db(
        DATA_DIR / "home/databases/wdentity_redirections.db", read_only=True
    )
    wdclasses = get_class_db(DATA_DIR / "home/databases/wdclasses.db", read_only=True)
    wdprops = get_prop_db(DATA_DIR / "home/databases/wdprops.db", read_only=True)

    wdclasses = wdclasses.cache()
    wdprops = wdprops.cache()
    wdredirections = wdredirections.cache()

    normalize_semtab2022(
        DATASET_DIR / "semtab2022_round1/HardTablesR1/DataSets/HardTablesR1",
        DATASET_DIR / "semtab2022_hardtable_r1",
        wdredirections,
        wdclasses,
        wdprops,
    )



from __future__ import annotations
import orjson


from sm.dataset import Dataset
from sm_datasets.helper import (
    ROOT_DIR,
)
from sm_datasets.datasets import Datasets
import random


def migrate():
    examples = Datasets().semtab2020r4()
    table_ids = set()
    for path in (ROOT_DIR / "semtab2020_r4sampled" / "tables").glob("*.json"):
        table_ids.add(path.stem)

    selected_examples = [e for e in examples if e.table.table.table_id in table_ids]
    assert len(selected_examples) == len(table_ids)

    Dataset(ROOT_DIR / "semtab2020_r4sampled").save(
        selected_examples, clean_previous_data=False, table_fmt_indent=2
    )


def sample():
    examples = Datasets().semtab2020r4()
    random.Random(72).shuffle(examples)
    selected_examples = [e for e in examples if e.table.table.shape()[0] <= 100][:4096]
    (ROOT_DIR / "semtab2020_round4" / "sampled_4k.json").write_bytes(
        orjson.dumps([e.table.table.table_id for e in selected_examples])
    )


if __name__ == "__main__":
    # migrate()
    sample()



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping, Set
from pathlib import Path
from typing import Any, Sequence

import orjson
import pandas as pd
from kgdata.models.entity import Entity
from kgdata.models.ont_property import OntologyProperty
from kgdata.wikidata.models.wdentity import WDEntity
from loguru import logger
from sm.dataset import Dataset, Example, FullTable
from sm.inputs.link import WIKIDATA, EntityId, Link
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.wikidata import WikidataNamespace
from sm.outputs.semantic_model import ClassNode, LiteralNode, LiteralNodeDataType

ROOT_DIR = Path(__file__).parent.parent.absolute()


class Datasets:
    def get_dataset(self, name: str) -> Dataset:
        return getattr(self, name)()

    def wt250(self, fix_el: bool = True):
        dataset = Dataset(ROOT_DIR / "250wt")
        examples = dataset.load()
        if fix_el:
            for file in (dataset.location / "el_corrections/tables").iterdir():
                if not file.name.endswith(".tsv"):
                    continue

                with open(file, "r") as stream:
                    table_id = orjson.loads(stream.readline())
                    table = [
                        ex for ex in examples if ex.table.table.table_id == table_id
                    ][0].table
                    df = pd.read_csv(
                        stream,
                        sep="\t",
                        dtype={
                            "url": str,
                            "row": int,
                            "col": int,
                            "start": int,
                            "end": int,
                            "entity": str,
                        },
                    )
                    df[["url"]] = df[["url"]].fillna("")
                    pos2rows = defaultdict(list)
                    for row in df.to_dict(orient="records"):
                        ri, ci = row["row"], row["col"]
                        pos2rows[ri, ci].append(row)

                    for (ri, ci), rows in pos2rows.items():
                        links = []
                        for row in rows:
                            link = Link(
                                start=row["start"],
                                end=row["end"],
                                url=row["url"],
                                entities=[EntityId(row["entity"], WIKIDATA)],
                            )
                            links.append(link)
                        table.links[ri, ci] = links
        return examples

    def semtab2022_r1(self):
        return Dataset(ROOT_DIR / "semtab2022_hardtable_r1")

    def semtab2019_t2dv2_dbpedia(self):
        return Dataset(ROOT_DIR / "semtab2019_t2dv2/dbpedia")

    def semtab2020r4(self):
        return Dataset(ROOT_DIR / "semtab2020_round4")

    def semtab2020r4_sampled50(self):
        return Dataset(ROOT_DIR / "semtab2020_r4sampled")

    def semtab2020r4_sampled512(self):
        examples = {e.table.table.table_id: e for e in self.semtab2020r4().load()}
        return [
            examples[eid]
            for eid in orjson.loads(
                (ROOT_DIR / "semtab2020_round4/sampled_4k.json").read_bytes()
            )[:512]
        ]

    def biotable(self):
        return Dataset(ROOT_DIR / "biotables")

    def biotable_rowsampled200(self):
        examples = {e.table.table.table_id: e for e in self.biotable().load()}
        for eid, sample in orjson.loads(
            (ROOT_DIR / "biotables" / "sampled_rows.json").read_bytes()
        ).items():
            examples[eid].table.keep_rows(sample[:200])
        return list(examples.values())

    def fix_redirection(
        self,
        examples: list[Example[FullTable]],
        entities: Mapping[str, Any] | Set[str],
        props: Mapping[str, OntologyProperty] | Set[str],
        redirections: Mapping[str, str],
        kgns: KnowledgeGraphNamespace,
        skip_unk_ont_ent: bool = False,
        skip_no_sm: bool = False,
    ):
        new_examples: list[Example[FullTable]] = []
        for example in examples:
            table = example.table
            for cell in table.links.flat_iter():
                for link in cell:
                    link.entities = self._fix_redirections(
                        link.entities, entities, redirections
                    )

            table.context.page_entities = self._fix_redirections(
                table.context.page_entities, entities, redirections
            )

            new_sms = []
            for sm_i, sm in enumerate(example.sms):
                skip_sm = False
                for n in sm.iter_nodes():
                    if isinstance(n, ClassNode):
                        assert kgns.is_uri(n.abs_uri)
                        if kgns.is_abs_uri_entity(n.abs_uri):
                            qid = kgns.get_entity_id(n.abs_uri)
                            if qid not in entities:
                                # if the qid not in redirection, the class is deleted, we should consider remove this example
                                new_qid = redirections[qid]
                                logger.debug("Redirect entity: {} to {}", qid, new_qid)

                                assert new_qid in entities, (
                                    "Just to be safe that qnodes & redirections are consistent",
                                    new_qid,
                                )
                                n.abs_uri = kgns.get_entity_abs_uri(new_qid)
                                n.rel_uri = kgns.get_entity_rel_uri(new_qid)
                    if isinstance(n, LiteralNode):
                        if n.datatype == LiteralNodeDataType.Entity:
                            assert kgns.is_uri(n.value)
                            qid = kgns.get_entity_id(n.value)
                            if qid not in entities:
                                # if the qid not in redirection, the class is deleted, we should consider remove this example
                                new_qid = redirections[qid]
                                logger.debug("Redirect entity: {} to {}", qid, new_qid)

                                assert new_qid in entities, (
                                    "Just to be safe that entities & redirections are consistent",
                                    new_qid,
                                )
                                n.value = WikidataNamespace.get_entity_abs_uri(new_qid)
                pid = None
                for e in sm.iter_edges():
                    assert kgns.is_uri(e.abs_uri)
                    if not kgns.is_abs_uri_property(e.abs_uri):
                        continue
                    pid = kgns.get_prop_id(e.abs_uri)
                    if pid not in props:
                        if pid in redirections:
                            new_pid = redirections[pid]
                            logger.debug("Redirect property: {} to {}", pid, new_pid)

                            assert new_pid in props, (
                                "Just to be safe that entities & redirections are consistent",
                                new_pid,
                            )
                            e.abs_uri = kgns.get_entity_abs_uri(new_pid)
                            e.rel_uri = kgns.get_entity_rel_uri(new_pid)
                        else:
                            if skip_unk_ont_ent:
                                # if the pid not in redirection, the property is deleted, we should consider remove this example
                                skip_sm = True
                                break
                            else:
                                raise KeyError("Unknown property", pid)

                if skip_sm and skip_unk_ont_ent:
                    logger.debug(
                        "Skip the semantic model at pos {} of example {} due to unknown property {}",
                        sm_i,
                        example.table.table.table_id,
                        pid,
                    )
                    continue
                new_sms.append(sm)

            example.sms = new_sms
            if len(example.sms) == 0 and skip_no_sm:
                logger.debug(
                    "Skip the example {} due to no semantic model",
                    example.table.table.table_id,
                )
                continue
            new_examples.append(example)

        return new_examples

    def _fix_redirections(
        self,
        entids: list[EntityId],
        entities: Mapping[str, Any] | Set[str],
        redirections: Mapping[str, str],
    ) -> list[EntityId]:
        newents = []
        for entid in entids:
            if entid not in entities:
                newid = redirections.get(entid, None)
                logger.debug("Redirect entity: {} to {}", entid, newid)
                assert newid is None or newid in entities, (
                    "Just to be safe that entities & redirections are consistent",
                    newid,
                )
                if newid is not None:
                    newents.append(EntityId(newid, entid.type))
            else:
                newents.append(entid)
        return newents


if __name__ == "__main__":
    # exs = Datasets().wt250()
    exs = Datasets().biotable().load()
    print(len(exs))



"""Converting SemTab datasets into sem-desc format."""
from __future__ import annotations

import csv
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import asdict, dataclass
from glob import glob
from io import StringIO
from pathlib import Path
from typing import List, Literal, Mapping, Optional, Tuple, Union, cast
from zipfile import ZipFile

import orjson
import pandas as pd
import serde.csv
import serde.json
import serde.textline
import yada
from kgdata.wikidata.db import (
    WikidataDB,
    get_class_db,
    get_entity_redirection_db,
    get_prop_db,
)
from kgdata.wikidata.models import WDClass, WDProperty
from loguru import logger
from rdflib import RDFS
from sm.dataset import Context, FullTable
from sm.inputs.link import WIKIDATA, EntityId, Link
from sm.misc.matrix import Matrix
from sm.namespaces.wikidata import WikidataNamespace
from sm.prelude import I, M, O
from tqdm import tqdm

PathOrStr = Union[str, Path]


@dataclass
class CEA:
    row: int
    col: int
    ent_uris: list[str]


@dataclass
class CTA:
    col: int
    type_uri: str


@dataclass
class CPA:
    source: int
    target: int
    prop_uri: str


class SemTab(ABC):
    def __init__(
        self,
        indir: Path,
        outdir: Path,
        ont: Ontology,
        unk_cta_handler: Literal["ignore", "skip", "error"],
    ):
        self.indir = indir
        self.outdir = outdir
        self.ont = ont
        self.unk_cta_handler = unk_cta_handler

    def normalize(
        self,
    ):
        """Normalize semtab2022 dataset. This function tested on HardTablesR1.

        The original semtab2022 dataset should have the following format:

        The output will have the following format:
        descriptions
        ├── part-<num>.zip
        │   ├── <table_file_name>.json
        tables
        ├── part-<num>.zip
        │   ├── <table_file_name>.json

        Args:
            indir: the original Round_<num> directory
            outdir: the output directory
        """
        table2cea = self.get_table2cea()
        table2cta = self.get_table2cta()
        table2cpa = self.get_table2cpa()

        outputs: List[Tuple[O.SemanticModel, FullTable]] = []

        # verifying that the target and ground truth are the same
        target_cea = self.get_target_cea()
        target_cta = self.get_target_cta()
        target_cpa = self.get_target_cpa()

        assert len(set(target_cea.keys()).symmetric_difference(table2cea.keys())) == 0
        assert len(set(target_cta.keys()).symmetric_difference(table2cta.keys())) == 0
        assert len(set(target_cpa.keys()).symmetric_difference(table2cpa.keys())) == 0

        for table_id, lst in table2cea.items():
            assert (
                set(target_cea[table_id]).symmetric_difference(
                    [(x.row, x.col) for x in lst]
                )
                == set()
            )

        for table_id, lst in table2cta.items():
            assert (
                set(target_cta[table_id]).symmetric_difference([x.col for x in lst])
                == set()
            )

        for table_id, lst in table2cpa.items():
            assert (
                set(target_cpa[table_id]).symmetric_difference(
                    [(x.source, x.target) for x in lst]
                )
                == set()
            )

        print("Number of tables:")
        print(
            "- CEA ground-truth: #tables =",
            len(table2cea),
            ", #entities =",
            sum(len(lst) for lst in table2cea.values()),
        )
        print(
            "- CTA ground-truth: #tables =",
            len(table2cta),
            ", #entity-columns =",
            sum(len(lst) for lst in table2cta.values()),
        )
        print(
            "- CPA ground-truth: #tables =",
            len(table2cpa),
            ", #relationships = ",
            sum(len(lst) for lst in table2cpa.values()),
        )

        # extract tables
        print("- # CSV File:", sum(1 for _ in (self.indir / "tables").glob("*.csv")))
        ignore_logs = []

        for file in tqdm(sorted((self.indir / "tables").glob("*.csv"))):
            table_id = file.stem
            if table_id not in table2cta:
                assert table_id not in table2cpa
                continue

            with open(file, mode="r") as f:
                reader = csv.reader(f, delimiter=",")
                lst = [row for row in reader]
                header, rows = lst[0], lst[1:]
                columns = []
                for ci in range(len(header)):
                    columns.append(
                        I.Column(
                            index=ci, name=header[ci], values=[row[ci] for row in rows]
                        )
                    )

                table = I.ColumnBasedTable(table_id=table_id, columns=columns)
                shp = table.shape()
                links = [[[] for ci in range(shp[1])] for ri in range(shp[0])]
                for cea in table2cea[table_id]:
                    links[cea.row][cea.col].append(
                        Link(
                            start=0,
                            end=len(table[cea.row, cea.col]),
                            url=None,
                            entities=[
                                self.ont.get_entity_id(uri) for uri in cea.ent_uris
                            ],
                        )
                    )

                sm = O.SemanticModel()
                col2node = {}
                for col in table.columns:
                    col2node[col.index] = sm.add_node(
                        O.DataNode(col.index, cast(str, col.name))
                    )

                assert table_id in table2cta
                rels = table2cpa.get(table_id, [])
                types = table2cta[table_id]
                ignore_table = False

                for cta in types:
                    try:
                        label = self.ont.get_class_readable_label(cta.type_uri)
                    except KeyError:
                        logger.error(
                            "Table {} contains unknown class {} in column {}",
                            table_id,
                            cta.type_uri,
                            cta.col,
                        )
                        label = None
                        if self.unk_cta_handler == "ignore":
                            ignore_table = True
                            ignore_logs.append(
                                f"Ignore table {table_id} as it contains unknown class {cta.type_uri} in column {cta.col}"
                            )
                            break
                        elif self.unk_cta_handler == "error":
                            raise

                    class_id = sm.add_node(
                        O.ClassNode(
                            abs_uri=self.ont.get_entity_abs_uri(cta.type_uri),
                            rel_uri=self.ont.get_entity_rel_uri(cta.type_uri),
                            approximation=False,
                            readable_label=label,
                        )
                    )
                    sm.add_edge(
                        O.Edge(
                            source=class_id,
                            target=col2node[cta.col],
                            abs_uri=str(RDFS.label),
                            rel_uri=self.ont.get_rel_uri(RDFS.label),
                            approximation=False,
                            readable_label=self.ont.get_rel_uri(RDFS.label),
                        )
                    )
                    col2node[cta.col] = class_id

                if ignore_table:
                    continue

                for cpa in rels:
                    pid = cpa.prop_uri
                    c1, c2 = cpa.source, cpa.target
                    try:
                        label = self.ont.get_prop_readable_label(cpa.prop_uri)
                    except KeyError:
                        logger.error(
                            "Table {} contains unknown property {} from column {} -> {}",
                            table_id,
                            cpa.prop_uri,
                            c1,
                            c2,
                        )
                        label = None
                        if self.unk_cta_handler == "ignore":
                            ignore_table = True
                            ignore_logs.append(
                                f"Ignore table {table_id} as it contains unknown property {cpa.prop_uri} from column {c1} -> {c2}"
                            )
                            break

                    source_node = sm.get_node(col2node[c1])
                    if not (
                        isinstance(source_node, O.ClassNode)
                        or (
                            isinstance(source_node, O.LiteralNode)
                            and source_node.datatype == O.LiteralNodeDataType.Entity
                        )
                    ):
                        ignore_table = True
                        ignore_logs.append(
                            f"Ignore table {table_id} as it contains edge from column {c1} -> {c2} but {c1} is not a class or entity node"
                        )
                        break
                    sm.add_edge(
                        O.Edge(
                            source=col2node[c1],
                            target=col2node[c2],
                            abs_uri=self.ont.get_prop_abs_uri(pid),
                            rel_uri=self.ont.get_prop_rel_uri(pid),
                            approximation=False,
                            readable_label=label,
                        )
                    )

                if ignore_table:
                    continue

                outputs.append(
                    (sm, FullTable(table=table, context=Context(), links=Matrix(links)))
                )

        serde.textline.ser(ignore_logs, self.outdir / "ignore_tables.log")

        outputs = sorted(outputs, key=lambda x: x[1].table.table_id)
        (self.outdir / "descriptions").mkdir(exist_ok=True, parents=True)
        (self.outdir / "tables").mkdir(exist_ok=True, parents=True)

        batch_size = 500

        # compress the content to save space, use zless, zcat to browse it
        serde.csv.ser(
            [
                [table_id, self.ser_semtab_row_index(row), str(col)]
                for table_id, lst in target_cea.items()
                for row, col in lst
            ],
            self.outdir / f"cea_targets.csv.gz",
        )
        serde.csv.ser(
            [
                [table_id, str(col)]
                for table_id, lst in target_cta.items()
                for col in lst
            ],
            self.outdir / f"cta_targets.csv.gz",
        )
        serde.csv.ser(
            [
                [table_id, str(source), str(target)]
                for table_id, lst in target_cpa.items()
                for source, target in lst
            ],
            self.outdir / f"cpa_targets.csv.gz",
        )

        counter = 0
        for i in range(0, len(outputs), batch_size):
            batch = outputs[i : i + batch_size]
            filename = f"part-{counter:04d}.zip"
            with ZipFile(self.outdir / "descriptions" / filename, "w") as zf:
                for sm, table in batch:
                    zf.writestr(
                        table.table.table_id + ".json",
                        data=orjson.dumps([sm.to_dict()]),
                    )
            with ZipFile(self.outdir / "tables" / filename, "w") as zf:
                for sm, table in batch:
                    zf.writestr(
                        table.table.table_id + ".json",
                        data=orjson.dumps(table.to_dict()),
                    )
            counter += 1

    def get_table2cea(self) -> dict[str, list[CEA]]:
        rows = serde.csv.deser(self.get_gt_cea_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 4
            table_id, ri, ci = row[0], int(row[1]) - 1, int(row[2])
            ent_uris = row[3].split(" ")
            output[table_id].append(CEA(row=ri, col=ci, ent_uris=ent_uris))
        return dict(output)

    def get_table2cta(self) -> dict[str, list[CTA]]:
        rows = serde.csv.deser(self.get_gt_cta_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 3, row
            table_id = row[0]
            output[table_id].append(CTA(col=int(row[1]), type_uri=row[2]))
        return dict(output)

    def get_table2cpa(self) -> dict[str, list[CPA]]:
        rows = serde.csv.deser(self.get_gt_cpa_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 4
            table_id = row[0]
            output[table_id].append(
                CPA(source=int(row[1]), target=int(row[2]), prop_uri=row[3])
            )
        return dict(output)

    def get_target_cea(self) -> dict[str, list[tuple[int, int]]]:
        rows = serde.csv.deser(self.get_target_cea_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 3
            table_id, ri, ci = row[0], self.deser_semtab_row_index(row[1]), int(row[2])
            output[table_id].append((ri, ci))
        return dict(output)

    def get_target_cta(self) -> dict[str, list[int]]:
        rows = serde.csv.deser(self.get_target_cta_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 2
            if row[0] == "" and row[1] == "":
                # encounter this error on semtab t2dv2
                continue
            table_id, col = row[0], int(row[1])
            output[table_id].append(col)
        return dict(output)

    def get_target_cpa(self) -> dict[str, list[tuple[int, int]]]:
        cpa_file = self.get_target_cpa_file()
        assert cpa_file is not None
        rows = serde.csv.deser(cpa_file)
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 2
            table_id, source, target = row[0], int(row[1]), int(row[2])
            output[table_id].append((source, target))
        return dict(output)

    def deser_semtab_row_index(self, row_index: str) -> int:
        # semtab row starts from 1 instead of 0
        return int(row_index) - 1

    def ser_semtab_row_index(self, row_index: int) -> str:
        # semtab row starts from 1 instead of 0
        return str(row_index + 1)

    @abstractmethod
    def get_gt_cea_file(self) -> Path:
        ...

    @abstractmethod
    def get_gt_cta_file(self) -> Path:
        ...

    @abstractmethod
    def get_gt_cpa_file(self) -> Path:
        ...

    @abstractmethod
    def get_target_cea_file(self) -> Path:
        ...

    @abstractmethod
    def get_target_cta_file(self) -> Path:
        ...

    @abstractmethod
    def get_target_cpa_file(self) -> Optional[Path]:
        ...


class SemTab2019(SemTab):
    def get_table2cea(self) -> dict[str, list[CEA]]:
        rows = serde.csv.deser(self.get_gt_cea_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 4
            # in semtab t2d dataset, column first, then row
            table_id, ci, ri = row[0], int(row[1]), int(row[2]) - 1
            ent_uris = row[3].split(" ")
            output[table_id].append(CEA(row=ri, col=ci, ent_uris=ent_uris))
        return dict(output)

    def get_target_cea(self) -> dict[str, list[tuple[int, int]]]:
        rows = serde.csv.deser(self.get_target_cea_file())
        output = defaultdict(list)
        for row in rows:
            assert len(row) == 3
            # in semtab t2d dataset, column first, then row
            table_id, ci, ri = row[0], int(row[1]), int(row[2]) - 1
            output[table_id].append((ri, ci))
        return dict(output)

    def get_target_cpa(self) -> dict[str, list[tuple[int, int]]]:
        if self.get_target_cpa_file() is None:
            assert (
                str(self.indir).find("t2dv2") != -1
            ), "t2dv2 dataset don't have CPA target file"
            # use the gt file as target
            return {
                table_id: [(x.source, x.target) for x in lst]
                for table_id, lst in self.get_table2cpa().items()
            }

        return super().get_target_cpa()

    def get_gt_cea_file(self) -> Path:
        # make sure only one file matches the pattern
        (infile,) = glob(str(self.indir / "gt/CEA_*_gt.csv"))
        return Path(infile)

    def get_gt_cta_file(self) -> Path:
        # make sure only one file matches the pattern
        (infile,) = glob(str(self.indir / "gt/CTA_*_gt.csv"))
        return Path(infile)

    def get_gt_cpa_file(self) -> Path:
        # make sure only one file matches the pattern
        (infile,) = glob(str(self.indir / "gt/CPA_*_gt.csv"))
        return Path(infile)

    def get_target_cea_file(self) -> Path:
        # make sure only one file matches the pattern
        (infile,) = glob(str(self.indir / "targets/CEA_*_Targets.csv"))
        return Path(infile)

    def get_target_cta_file(self) -> Path:
        # make sure only one file matches the pattern
        (infile,) = glob(str(self.indir / "targets/CTA_*_Targets.csv"))
        return Path(infile)

    def get_target_cpa_file(self) -> Optional[Path]:
        # make sure only one file matches the pattern
        infiles = glob(str(self.indir / "targets/CPA_*_Targets.csv"))
        if len(infiles) == 0:
            return None
        else:
            assert len(infiles) == 1
            return Path(infiles[0])


class Ontology(ABC):
    @abstractmethod
    def get_prop_readable_label(self, prop_uri: str) -> str:
        ...

    @abstractmethod
    def get_class_readable_label(
        self,
        class_uri: str,
    ) -> str:
        ...

    @abstractmethod
    def get_entity_id(self, ent_uri: str) -> EntityId:
        ...

    @abstractmethod
    def get_entity_abs_uri(self, ent_uri: str) -> str:
        ...

    @abstractmethod
    def get_entity_rel_uri(self, ent_uri: str) -> str:
        ...

    @abstractmethod
    def get_prop_abs_uri(self, prop_uri: str) -> str:
        ...

    @abstractmethod
    def get_prop_rel_uri(self, prop_uri: str) -> str:
        ...

    @abstractmethod
    def get_rel_uri(self, uri: str) -> str:
        ...


class WikidataOntology(Ontology):
    def __init__(self, db: WikidataDB):
        self.db = db
        self.wdns = WikidataNamespace.create()

    def get_entity_id(self, ent_uri: str) -> EntityId:
        return EntityId(self.wdns.get_entity_id(ent_uri), WIKIDATA)

    def get_prop_readable_label(
        self,
        prop_uri: str,
    ):
        pid = self.wdns.get_prop_id(prop_uri)
        if pid in self.db.redirections:
            pid = self.db.redirections[pid]

        return f"{self.db.props[pid].label} ({pid})"

    def get_class_readable_label(
        self,
        class_uri: str,
    ):
        eid = self.wdns.get_entity_id(class_uri)
        if eid in self.db.redirections:
            eid = self.db.redirections[eid]

        return f"{self.db.classes[eid].label} ({eid})"

    def get_entity_abs_uri(self, ent_uri: str) -> str:
        return self.wdns.get_entity_abs_uri(self.wdns.get_entity_id(ent_uri))

    def get_entity_rel_uri(self, ent_uri: str) -> str:
        return self.wdns.get_entity_rel_uri(self.wdns.get_entity_id(ent_uri))

    def get_prop_abs_uri(self, prop_uri: str) -> str:
        return self.wdns.get_prop_abs_uri(self.wdns.get_prop_id(prop_uri))

    def get_prop_rel_uri(self, prop_uri: str) -> str:
        return self.wdns.get_prop_rel_uri(self.wdns.get_prop_id(prop_uri))

    def get_rel_uri(self, uri: str) -> str:
        return self.wdns.get_rel_uri(uri)


class DBpediaOntology(Ontology):
    def get_prop_readable_label(self, prop_uri: str) -> str:
        return prop_uri

    def get_class_readable_label(
        self,
        class_uri: str,
    ) -> str:
        return class_uri

    def get_entity_id(self, ent_uri: str) -> EntityId:
        return EntityId(ent_uri, "dbpedia")

    def get_entity_abs_uri(self, ent_uri: str) -> str:
        return ent_uri

    def get_entity_rel_uri(self, ent_uri: str) -> str:
        return ent_uri

    def get_prop_abs_uri(self, prop_uri: str) -> str:
        return prop_uri

    def get_prop_rel_uri(self, prop_uri: str) -> str:
        return prop_uri

    def get_rel_uri(self, uri: str) -> str:
        return uri


@dataclass
class CLI:
    database_dir: Path
    input_dir: Path
    output_dir: Path

    ontology: Literal["wikidata", "dbpedia"]
    unk_cta_handler: Literal["ignore", "skip", "error"] = "error"


if __name__ == "__main__":
    cli = yada.Parser1(CLI).parse_args()

    if cli.ontology == "wikidata":
        ont = WikidataOntology(WikidataDB(cli.database_dir))
    else:
        assert cli.ontology == "dbpedia"
        ont = DBpediaOntology()

    if not cli.output_dir.exists():
        cli.output_dir.mkdir(parents=True)
        metadata = {
            "ontology": cli.ontology,
            "unk_cta_handler": cli.unk_cta_handler,
        }
        if (cli.output_dir / "metadata.json").exists():
            assert metadata == serde.json.deser(cli.output_dir / "metadata.json")
        else:
            serde.json.ser(metadata, cli.output_dir / "metadata.json")
    SemTab2019(cli.input_dir, cli.output_dir, ont, cli.unk_cta_handler).normalize()



"""NOTE: this script is copied from a local folder and hasn't been re-tested. For now it serves as
a reference how semtab2020 dataset is converted
"""

import csv
from collections import defaultdict
from io import StringIO
from pathlib import Path
from typing import List, Mapping, Optional, Tuple, Union, cast
from zipfile import ZipFile

import orjson
import serde.csv
from kgdata.wikidata.db import get_class_db, get_entity_redirection_db, get_prop_db
from kgdata.wikidata.models import WDClass, WDProperty
from loguru import logger
from rdflib import RDFS
from sm.namespaces.wikidata import WikidataNamespace
from sm.prelude import I, M, O
from tqdm import tqdm

from grams.inputs.linked_table import Context, Link, LinkedTable
from scripts.config import DATA_DIR, DATASET_DIR

PathOrStr = Union[str, Path]


def normalize_semtab2020(
    indir: PathOrStr,
    outdir: PathOrStr,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    """Normalize semtab2020 dataset. This function works for data on each round.

    The original semtab2020 dataset should have the following format:

    GT
    ├── CEA
    │   ├── CEA_Round<num>_gt.csv
    ├── CPA
    │   ├── CPA_Round<num>_gt.csv
    ├── CTA
    │   ├── CTA_Round<num>_gt.csv
    Round<num>
    ├── CEA_Round<num>_targets.csv
    ├── CPA_Round<num>_targets.csv
    ├── CTA_Round<num>_targets.csv
    ├── Tables_Round<num>.zip

    The output will have the following format:
    descriptions
    ├── part-<num>.zip
    │   ├── <table_file_name>.json
    tables
    ├── part-<num>.zip
    │   ├── <table_file_name>.json

    Args:
        indir: the original Round_<num> directory
        outdir: the output directory
    """
    indir = Path(indir)
    outdir = Path(outdir)

    # round number
    no = int(indir.name.replace("Round", ""))

    cea_gt = indir.parent / f"GT/CEA/CEA_Round{no}_gt.csv"
    cpa_gt = indir.parent / f"GT/CPA/CPA_Round{no}_gt.csv"
    cta_gt = indir.parent / f"GT/CTA/CTA_Round{no}_gt.csv"

    cea = serde.csv.deser(cea_gt)
    cpa = serde.csv.deser(cpa_gt)
    cta = serde.csv.deser(cta_gt)

    # gather links and prop/types
    table2links = defaultdict(list)
    table2sem = defaultdict(lambda: ([], []))

    for row in cea:
        table2links[row[0]].append(row)
    for row in cpa:
        table2sem[row[0]][0].append(row)
    for row in cta:
        table2sem[row[0]][1].append(row)

    wdns = WikidataNamespace.create()
    outputs: List[Tuple[O.SemanticModel, LinkedTable]] = []

    cea_tables = {row[0] for row in cea}
    cta_tables = {row[0] for row in cta}

    # verifying that the target and ground truth are the same
    for i, row in enumerate(serde.csv.deser(indir / f"CEA_Round{no}_targets.csv")):
        if i >= len(cea) or row != cea[i][:3]:
            assert (
                row[0] not in cea_tables
            ), f"when a target entity is not in ground-truth. the whole table {row[0]} must not be in the ground-truth"
    for i, row in enumerate(serde.csv.deser(indir / f"CPA_Round{no}_targets.csv")):
        assert row == cpa[i][:3]
    for i, row in enumerate(serde.csv.deser(indir / f"CTA_Round{no}_targets.csv")):
        if i >= len(cta) or row != cta[i][:2]:
            assert (
                row[0] not in cta_tables
            ), f"when a target column is not in ground-truth. the whole table {row[0]} must not be in the ground-truth"

    print("Number of tables:")
    print("- CEA ground-truth:", len(cea_tables))
    print("- CTA ground-truth:", len(set(row[0] for row in cta)))
    print("- CPA ground-truth:", len(cta_tables))

    # extract tables
    with ZipFile(indir / f"Tables_Round{no}.zip") as zf:
        print("- CSV File:", len(zf.infolist()))
        ignore_logs = []

        for file in tqdm(zf.infolist()):
            if not file.filename.endswith(".csv"):
                continue
            table_id = Path(file.filename).stem
            with zf.open(file, mode="r") as f:
                reader = csv.reader(StringIO(f.read().decode()), delimiter=",")
                lst = [row for row in reader]
                header, rows = lst[0], lst[1:]
                columns = []
                for ci in range(len(header)):
                    columns.append(
                        I.Column(
                            index=ci, name=header[ci], values=[row[ci] for row in rows]
                        )
                    )

                table = I.ColumnBasedTable(table_id=table_id, columns=columns)
                shp = table.shape()
                links = [[[] for ci in range(shp[1])] for ri in range(shp[0])]
                for row in table2links[table_id]:
                    ri, ci = int(row[1]) - 1, int(row[2])
                    ent_uri = row[-1]
                    links[ri][ci].append(
                        Link(
                            start=0,
                            end=len(table[ri, ci]),
                            url=None,
                            entity_id=wdns.get_entity_id(ent_uri),
                            candidates=[],
                        )
                    )

                sm = O.SemanticModel()
                col2node = {}
                for col in table.columns:
                    col2node[col.index] = sm.add_node(
                        O.DataNode(col.index, cast(str, col.name))
                    )

                rels, types = table2sem[table_id]
                ignore_table = False

                for row in types:
                    cid = int(row[1])
                    ent_id = wdns.get_entity_id(row[-1])

                    try:
                        label = get_class_readable_label(
                            ent_id, wdredirections, wdclasses
                        )
                    except KeyError:
                        logger.error(
                            "Table {} contains unknown class {} in column {}",
                            table_id,
                            ent_id,
                            cid,
                        )
                        label = None
                        if wdclasses is not None and wdredirections is not None:
                            ignore_table = True
                            ignore_logs.append(
                                f"Ignore table {table_id} as it contains unknown class {ent_id} in column {cid}"
                            )
                            break

                    class_id = sm.add_node(
                        O.ClassNode(
                            abs_uri=wdns.get_entity_abs_uri(ent_id),
                            rel_uri=wdns.get_entity_rel_uri(ent_id),
                            approximation=False,
                            readable_label=label,
                        )
                    )
                    sm.add_edge(
                        O.Edge(
                            source=class_id,
                            target=col2node[cid],
                            abs_uri=str(RDFS.label),
                            rel_uri=wdns.get_rel_uri(RDFS.label),
                            approximation=False,
                            readable_label=wdns.get_rel_uri(RDFS.label),
                        )
                    )
                    col2node[cid] = class_id

                if ignore_table:
                    continue

                for row in rels:
                    pid = wdns.get_prop_id(row[-1])
                    c1, c2 = int(row[1]), int(row[2])
                    try:
                        label = get_prop_readable_label(pid, wdredirections, wdprops)
                    except KeyError:
                        logger.error(
                            "Table {} contains unknown property {} from column {} -> {}",
                            table_id,
                            pid,
                            c1,
                            c2,
                        )
                        label = None
                        if wdprops is not None and wdredirections is not None:
                            ignore_table = True
                            ignore_logs.append(
                                f"Ignore table {table_id} as it contains unknown property {pid} from column {c1} -> {c2}"
                            )
                            break

                    sm.add_edge(
                        O.Edge(
                            source=col2node[c1],
                            target=col2node[c2],
                            abs_uri=wdns.get_prop_abs_uri(pid),
                            rel_uri=wdns.get_prop_rel_uri(pid),
                            approximation=False,
                            readable_label=label,
                        )
                    )

                if ignore_table:
                    continue

                outputs.append(
                    (sm, LinkedTable(table=table, context=Context(), links=links))
                )

        M.serialize_lines(ignore_logs, outdir / "ignore_tables.log")

    outputs = sorted(outputs, key=lambda x: x[1].id)
    (outdir / "descriptions").mkdir(exist_ok=True, parents=True)
    (outdir / "tables").mkdir(exist_ok=True, parents=True)

    batch_size = 500

    for name in ["CEA", "CPA", "CTA"]:
        # compress the content to save space, use zless, zcat to browse it
        content = (indir / f"{name}_Round{no}_targets.csv").read_bytes()
        (outdir / f"{name}_targets.csv.gz").write_bytes(content)

    counter = 0
    for i in range(0, len(outputs), batch_size):
        batch = outputs[i : i + batch_size]
        filename = f"part-{counter:04d}.zip"
        with ZipFile(outdir / "descriptions" / filename, "w") as zf:
            for sm, table in batch:
                zf.writestr(table.id + ".json", data=orjson.dumps([sm.to_dict()]))
        with ZipFile(outdir / "tables" / filename, "w") as zf:
            for sm, table in batch:
                zf.writestr(table.id + ".json", data=orjson.dumps(table.to_dict()))
        counter += 1


def get_prop_readable_label(
    pid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    if wdprops is None:
        return None

    if wdredirections is not None and pid in wdredirections:
        pid = wdredirections[pid]

    return f"{wdprops[pid].label} ({pid})"


def get_class_readable_label(
    eid: str,
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
):
    if wdclasses is None:
        return None

    if wdredirections is not None and eid in wdredirections:
        eid = wdredirections[eid]

    return f"{wdclasses[eid].label} ({eid})"


if __name__ == "__main__":
    wdredirections = get_entity_redirection_db(
        DATA_DIR / "home/databases/wdentity_redirections.db", read_only=True
    )
    wdclasses = get_class_db(DATA_DIR / "home/databases/wdclasses.db", read_only=True)
    wdprops = get_prop_db(DATA_DIR / "home/databases/wdprops.db", read_only=True)

    wdclasses = wdclasses.cache()
    wdprops = wdprops.cache()
    wdredirections = wdredirections.cache()

    normalize_semtab2020(
        DATASET_DIR / "raw_semtab2020/Round4",
        DATASET_DIR / "semtab2020_round4",
        wdredirections,
        wdclasses,
        wdprops,
    )



from __future__ import annotations
from dataclasses import dataclass
from typing import List

from sm.inputs.prelude import ColumnBasedTable, Context, Link
from sm.misc.matrix import Matrix


@dataclass
class Table:
    table: ColumnBasedTable
    context: Context
    links: Matrix[List[Link]]

    def to_dict(self):
        return {
            "version": 2,
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in cell] for cell in row]
                for row in self.links.data
            ],
        }

    @staticmethod
    def from_dict(obj: dict):
        version = obj["version"]
        if not (version == "1.2" or version == 2):
            raise ValueError(f"Unknown version: {version}")

        return Table(
            table=ColumnBasedTable.from_dict(obj["table"]),
            context=Context.from_dict(obj["context"]),
            links=Matrix(
                [
                    [[Link.from_dict(link) for link in cell] for cell in row]
                    for row in obj["links"]
                ]
            ),
        )



from __future__ import annotations

from collections import defaultdict
import csv
from io import StringIO
from typing import Dict, List, Mapping, Optional, cast
from zipfile import ZipFile
from kgdata.wikidata.db import (
    WikidataDB,
)
from kgdata.wikidata.models import WDClass, WDProperty
from loguru import logger
import orjson
from rdflib import RDFS

from sm.dataset import Dataset, Example, FullTable
import sm.inputs.prelude as I
import sm.outputs.semantic_model as O
from sm.misc.matrix import Matrix
from sm.namespaces.wikidata import WikidataNamespace
from sm_datasets.helper import (
    DB_DIR,
    ROOT_DIR,
    get_class_readable_label,
    get_prop_readable_label,
)
import random
from tqdm import tqdm
from pathlib import Path


def normalize_biotables(
    dataset_name="biotables",
    wdredirections: Optional[Mapping[str, str]] = None,
    wdclasses: Optional[Mapping[str, WDClass]] = None,
    wdprops: Optional[Mapping[str, WDProperty]] = None,
):
    ds = Dataset(ROOT_DIR / dataset_name)
    file = ds.location / "raw_data" / "BioTable-Datasets.zip"
    wdns = WikidataNamespace.create()
    ignore_logs = []
    ignore_tables = set()

    with ZipFile(file) as zf:
        tables: Dict[str, I.ColumnBasedTable] = {}
        table2links: Dict[str, Matrix[List[I.Link]]] = {}
        table2sm: Dict[str, O.SemanticModel] = {}

        gt_tasks = defaultdict(list)
        for file in tqdm(zf.infolist(), desc="parsing tables"):
            if file.filename.startswith("datasets"):
                table_id = Path(file.filename).stem
                with zf.open(file, mode="r") as f:
                    reader = csv.reader(StringIO(f.read().decode()), delimiter=",")
                    rows = [row for row in reader]
                    tables[table_id] = I.ColumnBasedTable.from_rows(
                        rows, table_id, strict=True
                    )
            else:
                assert file.filename.startswith("ground-truth"), file.filename
                assert file.filename.endswith(".csv")

                task = Path(file.filename).name[:3]
                gt_tasks[task].append(file)

        assert set(gt_tasks.keys()) == {"cea", "cpa", "cta"}
        gt_tasks = {k: gt_tasks[k] for k in ["cta", "cpa", "cea"]}
        table2col2node = {}
        for task, files in gt_tasks.items():
            gt, gt_target = sorted(files, key=lambda f: f.filename)
            with zf.open(gt, mode="r") as f:
                reader = csv.reader(StringIO(f.read().decode()), delimiter=",")
                gt_rows = [row for row in reader]
            with zf.open(gt_target, mode="r") as f:
                reader = csv.reader(StringIO(f.read().decode()), delimiter=",")
                gt_target_rows = [row for row in reader]

            assert len(gt_rows) == len(gt_target_rows)
            assert all(
                (x[0], x[-1]) == (y[0], y[-1]) for x, y in zip(gt_rows, gt_target_rows)
            )

            table2rows = defaultdict(list)
            for row in tqdm(gt_rows, desc=f"parsing {task} (step 1)"):
                table2rows[row[0]].append(row[1:])
            for tid, rows in tqdm(table2rows.items(), desc=f"parsing {task} (step 2)"):
                table = tables[tid]
                if task == "cea":
                    table2links[tid] = Matrix.default(table.shape(), lambda: list())
                    for ri, ci, wdenturl in rows:
                        table2links[tid][int(ri), int(ci)].append(
                            I.Link(
                                start=0,
                                end=len(table[int(ri), int(ci)]),
                                url=wdenturl,
                                entities=[
                                    I.EntityId(wdns.get_entity_id(wdenturl), "wikidata")
                                ],
                            )
                        )
                elif task == "cta":
                    assert tid not in table2sm
                    sm = O.SemanticModel(check_cycle=True, multigraph=False)
                    table2sm[tid] = sm

                    col2node = {}
                    for col in table.columns:
                        col2node[col.index] = sm.add_node(
                            O.DataNode(col.index, cast(str, col.name))
                        )

                    for ci, wdenturl in rows:
                        ci = int(ci)
                        entid = wdns.get_entity_id(wdenturl)

                        try:
                            label = get_class_readable_label(
                                entid, wdredirections, wdclasses
                            )
                        except KeyError:
                            logger.error(
                                "Table {} contains unknown class {} in column {}",
                                tid,
                                entid,
                                ci,
                            )
                            label = None
                            if wdclasses is not None and wdredirections is not None:
                                ignore_tables.add(tid)
                                ignore_logs.append(
                                    f"Ignore table {tid} as it contains unknown class {entid} in column {ci}"
                                )
                                break

                        class_id = sm.add_node(
                            O.ClassNode(
                                abs_uri=wdns.get_entity_abs_uri(entid),
                                rel_uri=wdns.get_entity_rel_uri(entid),
                                approximation=False,
                                readable_label=label,
                            )
                        )
                        sm.add_edge(
                            O.Edge(
                                source=class_id,
                                target=col2node[ci],
                                abs_uri=str(RDFS.label),
                                rel_uri=wdns.get_rel_uri(RDFS.label),
                                approximation=False,
                                readable_label=wdns.get_rel_uri(RDFS.label),
                            )
                        )
                        col2node[ci] = class_id

                    table2col2node[tid] = col2node
                elif task == "cpa":
                    assert tid in table2sm
                    sm = table2sm[tid]
                    col2node = table2col2node[tid]

                    for ci, cj, wdpropurl in rows:
                        pid = wdns.get_prop_id(wdpropurl)
                        label = get_prop_readable_label(pid, wdredirections, wdprops)
                        sm.add_edge(
                            O.Edge(
                                source=col2node[int(ci)],
                                target=col2node[int(cj)],
                                abs_uri=wdns.get_prop_abs_uri(pid),
                                rel_uri=wdns.get_prop_rel_uri(pid),
                                approximation=False,
                                readable_label=label,
                            )
                        )

        examples = []
        for tid, table in tables.items():
            if tid in ignore_tables:
                ignore_logs.append("Ignore table {tid}")
                continue

            examples.append(
                Example(
                    sms=[table2sm[tid]],
                    table=FullTable(table, context=I.Context(), links=table2links[tid]),
                )
            )

        ds.save(examples, individual_table_compressed="gz")


def sample_rows(dataset_name="biotables"):
    ds = Dataset(ROOT_DIR / dataset_name)
    if not (ds.location / "sampled_rows.json").exists():
        random.seed(72)
        output = {}
        for example in ds.load():
            nrows, ncols = example.table.table.shape()
            index = list(range(nrows))[:1000]
            random.shuffle(index)
            output[example.table.table.table_id] = index

        (ds.location / "sampled_rows.json").write_bytes(orjson.dumps(output))


if __name__ == "__main__":
    db = WikidataDB(DB_DIR)
    # normalize_biotables(
    #     wdredirections=db.wdredirections.cache(),
    #     wdclasses=db.wdclasses.cache(),
    #     wdprops=db.wdprops.cache(),
    # )
    sample_rows()
